module.exports = [
  {
    text: "Guide",
    link: "/guide/"
  },
  {
    text: "Reference",
    link: "/reference/"
  },
  {
    text: "Learn More",
    ariaLabel: "Learn More",
    items: [
      {
        text: "Adoption",
        items: [
          {
            text: "Migrating to machinable",
            link: "/miscellaneous/migration-guide.html"
          },
          // {
          //   text: "Best practices",
          //   link: "/miscellaneous/best-practice.html"
          // },
          {
            text: "FAQ",
            link: "/faq/"
          },
          // {
          //   text: "App",
          //   link: "/miscellaneous/app.html"
          // },
          {
            text: "Changelog",
            link: "https://github.com/machinable-org/machinable/blob/master/CHANGELOG.md"
          }
        ]
      },
      {
        text: "Meta",
        items: [
          // {
          //   text: "Design principles",
          //   link: "/miscellaneous/design-principles.html"
          // },
          // {
          //   text: "Comparison with alternatives",
          //   link: "/miscellaneous/comparison.html"
          // },
          {
            text: "Contribution guide",
            link: "/miscellaneous/contribution-guide.html"
          },
          // {
          //   text: "Glossary",
          //   link: "/miscellaneous/glossary.html"
          // }
        ]
      }
    ]
  }
];
