// placeholder for test, dont't remove it.

//.content {
//  font-size 30px;
//}

pre.vue-container
  border-left-width: .5rem;
  border-left-style: solid;
  border-color: #42b983;
  border-radius: 0px;
  & > code
    font-size: 14px !important;
    & > p
      margin: -5px 0 -20px 0;
    code
      background-color: #42b983 !important;
      padding: 3px 5px;
      border-radius: 3px;
      color #000
    em
      color #808080
      font-weight light

